@@ -1,10 +1,10 @@
 package Net::Hadoop::WebHDFS::LWP;
-
+$Net::Hadoop::WebHDFS::LWP::VERSION = '0.009';
 use strict;
 use warnings;
 use parent 'Net::Hadoop::WebHDFS';
 
-our $VERSION = '0.008'; # VERSION
+# VERSION
 
 use LWP::UserAgent;
 use Carp;
@@ -12,35 +12,42 @@ use Ref::Util    qw( is_arrayref );
 use Scalar::Util qw( openhandle );
 use HTTP::Request::StreamingUpload;
 
+use constant UA_PASSTHROUGH_OPTIONS => qw(
+    env_proxy
+    no_proxy
+    proxy
+);
+
 sub new {
     my $class   = shift;
     my %options = @_;
-    my $debug   = $options{debug} || 0;
+    my $debug   = delete $options{debug} || 0;
+
     require Data::Dumper if $debug;
 
     my $self = $class->SUPER::new(@_);
 
     # we don't need Furl
     delete $self->{furl};
+
     $self->{debug} = $debug;
 
-    my %ua_opts;
-    for my $passthru_opt (qw/ env_proxy proxy no_proxy /) {
-        $ua_opts{$passthru_opt} = $options{$passthru_opt}
-            if ( exists $options{$passthru_opt} );
-    }
-
-    $self->{ua} = LWP::UserAgent->new( %ua_opts );
-    $self->{ua}->agent("Net::Hadoop::WebHDFS::LWP " . $class->VERSION );
-    $self->{useragent} = $self->{ua}->agent;
-
     # default timeout is a bit short, raise it
-    $self->{timeout} = $options{timeout} || 30;
-    $self->{ua}->timeout( $self->{timeout} );
+    $self->{timeout}   = $options{timeout}   || 30;
 
     # For filehandle upload support
     $self->{chunksize} = $options{chunksize} || 4096;
 
+    $self->{ua_opts} = {
+        map {
+            exists $options{$_} ? (
+                $_ => $options{ $_ }
+            ) : ()
+        } UA_PASSTHROUGH_OPTIONS
+    };
+
+    $self->_create_ua;
+
     return $self;
 }
 
@@ -59,7 +66,7 @@ sub request {
     $uri->port($port);
     $uri->scheme('http'); # no ssl for webhdfs? check the docs
 
-    print "URI : $uri\n" if $self->{debug};
+    printf STDERR "URI : %s\n", $uri if $self->{debug};
 
     my $req;
 
@@ -89,10 +96,13 @@ sub request {
 
     my $res = { code => $real_res->code, body => $real_res->decoded_content };
     my $code = $real_res->code;
-    print "HTTP code : $code\n" if $self->{debug};
+
+    printf STDERR "HTTP code : %s\n", $code if $self->{debug};
 
     my $headers = $real_res->headers;
-    print "Headers: " . Data::Dumper::Dumper $headers if $self->{debug};
+
+    printf STDERR "Headers: %s", Data::Dumper::Dumper $headers if $self->{debug};
+
     for my $h_key ( keys %{ $headers || {} } ) {
         my $h_value = $headers->{$h_key};
 
@@ -106,23 +116,38 @@ sub request {
     my $errmsg = $res->{body} || 'Response body is empty...';
     $errmsg =~ s/\n//g;
 
-    if ( $code == 400 ) { croak "ClientError: $errmsg"; }
+    # Attempt to strigfy the HTML message
+    if ( $errmsg =~ m{ \A <html.+?> }xmsi ) {
+        if ( my @errors = $self->_parse_error_from_html( $errmsg ) ) {
+            # @error can also be assigned to a hash as it is mapped
+            # to kay=>value pairs, however strigifying the message
+            # is enough for now
+            my @flat;
+            while ( my ( $key, $val ) = splice( @errors, 0, 2 ) ) {
+                push @flat, "$key: $val"
+            }
+            # reset to something meaningful now that we've removed the html cruft
+            $errmsg = join '. ', @flat;
+        }
+    }
 
-    # this error happens for secure clusters when using Net::Hadoop::WebHDFS,
-    # but LWP::Authen::Negotiate takes care of it transparently in this module.
-    # we still may get this error on a secure cluster, when the credentials
-    # cache hasn't been initialized
+    if ( $code == 400 ) {
+        croak "ClientError: $errmsg";
+    }
     elsif ( $code == 401 ) {
+        # this error happens for secure clusters when using Net::Hadoop::WebHDFS,
+        # but LWP::Authen::Negotiate takes care of it transparently in this module.
+        # we still may get this error on a secure cluster, when the credentials
+        # cache hasn't been initialized
         my $extramsg = ( $headers->{'www-authenticate'} || '' ) eq 'Negotiate'
-            ? eval { require "LWP::Authen::Negotiate" }
-                ? ' (Did you forget to run kinit?)'
-                : ' (LWP::Authen::Negotiate doesn\'t seem available)'
+            ? eval { require LWP::Authen::Negotiate; 1; }
+                ? q{ (Did you forget to run kinit?)}
+                : q{ (LWP::Authen::Negotiate doesn't seem available)}
             : '';
         croak "SecurityError$extramsg: $errmsg";
     }
-
     elsif ( $code == 403 ) {
-        if ( $errmsg =~ /org\.apache\.hadoop\.ipc\.StandbyException/ ) {
+        if ( $errmsg =~ m{ \Qorg.apache.hadoop.ipc.StandbyException\E }xms ) {
             if ( $self->{httpfs_mode} || not defined( $self->{standby_host} ) ) {
 
                 # failover is disabled
@@ -135,27 +160,110 @@ sub request {
             else {
                 $self->{under_failover} = not $self->{under_failover};
                 $self->{retrying}       = 1;
-                my ( $next_host, $next_port ) = $self->connect_to();
-                my $val = $self->request( $next_host, $next_port, $method, $path, $op, $params,
-                    $payload, $header );
+                my ( $next_host, $next_port ) = $self->connect_to;
+                my $val = $self->request(
+                                $next_host,
+                                $next_port,
+                                $method,
+                                $path,
+                                $op,
+                                $params,
+                                $payload,
+                                $header,
+                            );
                 $self->{retrying} = 0;
                 return $val;
             }
         }
         croak "IOError: $errmsg";
     }
-    elsif ( $code == 404 ) { croak "FileNotFoundError: $errmsg"; }
-    elsif ( $code == 500 ) { croak "ServerError: $errmsg"; }
+    elsif ( $code == 404 ) {
+        croak "FileNotFoundError: $errmsg";
+    }
+    elsif ( $code == 500 ) {
+        croak "ServerError: $errmsg";
+    }
+    else {
+        # do nothing
+    }
 
+    # catch-all exception
     croak "RequestFailedError, code:$code, message:$errmsg";
 }
 
+sub _create_ua {
+    my $self  = shift;
+    my $class = ref $self;
+
+    $self->{ua} = LWP::UserAgent->new(
+                        %{ $self->{ua_opts} }
+                    );
+
+    $self->{ua}->agent(
+        sprintf "%s %s",
+                    $class,
+                    $class->VERSION || 'beta',
+    );
+
+    $self->{useragent} = $self->{ua}->agent;
+    $self->{ua}->timeout( $self->{timeout} );
+
+    return $self;
+}
+
+sub _parse_error_from_html {
+    # This is a brittle function as it assumes certain things to be present
+    # in the HTML output and will most likely break with future updates.
+    # However the interface returns HTML in certain cases (like secure clusters)
+    # and currently that's a failure on the backend where we can;t fix things.
+    #
+    # In any case, the program should default to the original message fetched,
+    # if this fails for any reason.
+    #
+    my $self   = shift;
+    my $errmsg = shift;
+
+    if ( ! eval { require HTML::Parser;} ) {
+        if ( $self->{debug} ) {
+            printf STDERR "Tried to parse the HTML error message but HTML::Parser is not available!\n";
+        }
+        return;
+    }
+
+    my @errors;
+    my $p = HTML::Parser->new(
+                api_version => 3,
+                handlers    => {
+                    text => [
+                        \@errors,
+                        'event,text',
+                    ],
+                }
+            );
+    $p->parse( $errmsg );
+
+    my @flat =  map {;
+                    s{ \A \s+    }{}xmsg;
+                    s{    \s+ \z }{}xmsg;
+                    $_;
+                }
+                grep {
+                       $_ !~ m{ \Q<!--\E    }xms # comment
+                    && $_ !~ m{ \A Apache\b }xms # Tomcat version, etc.
+                    && $_ !~ m{ \A \s+ \z   }xms # " "
+                }
+                map { $_->[1] }
+                @errors;
+
+    if ( @flat % 2 ) {
+        unshift @flat, 'http_status';
+    }
+
+    return @flat;
+}
+
 1;
 
-#ABSTRACT: Client library for Hadoop WebHDFS and HttpFs, with Kerberos support
-
-__END__
-
 =pod
 
 =encoding UTF-8
@@ -166,7 +274,7 @@ Net::Hadoop::WebHDFS::LWP - Client library for Hadoop WebHDFS and HttpFs, with K
 
 =head1 VERSION
 
-version 0.008
+version 0.009
 
 =head1 SYNOPSIS
 
@@ -199,6 +307,8 @@ As mentioned above, the real work was done by Satoshi Tagomori
 
 Thanks to my employer Booking.com to allow me to release this module for public use
 
+=for Pod::Coverage request
+
 =head1 AUTHOR
 
 David Morel <david.morel@amakuru.net>
@@ -211,3 +321,8 @@ This is free software; you can redistribute it and/or modify it under
 the same terms as the Perl 5 programming language system itself.
 
 =cut
+
+__END__
+
+#ABSTRACT: Client library for Hadoop WebHDFS and HttpFs, with Kerberos support
+
